from sky.configs import DEFAULT_CRAWL_CONFIG
from sky.scraper import Scrape

# Crawling 
CRAWL_CONFIG = DEFAULT_CRAWL_CONFIG
CRAWL_CONFIG.update({ 
    'seed_urls' : [ 
        'http://www.techcrunch.com/2015/05/21/chromebook-sales-predicted-to-grow-by-27-this-year-to-7-3m-units/'
    ],
    
    'collections_path' : '/Users/pascal/GDrive/sky_collections',

    'collection_name' : 'techie',

    # Optional
    
    'crawl_filter_regexps' : [ 
        
    ],

    'crawl_required_regexps' : [
        '2015', '2014'
    ],        
            
    'index_filter_regexps' : [
        
    ],
    
    'index_required_regexps' : [ 
        '2015', '2014'
    ], 

    'max_saved_responses' : 100,

    'max_workers' : 10,
})

from sky.crawler import crawl
crawl.start(CRAWL_CONFIG)

# Indexing

SCRAPE_CONFIG = CRAWL_CONFIG.copy()

SCRAPE_CONFIG.update({ 
    'template_proportion' : 0.09,
    'max_templates' : 1000
})

skindex = Scrape(SCRAPE_CONFIG)

res = skindex.process_all(remove_visuals = True)




import asyncio
from contextlib import contextmanager
import io
import logging
import socket
import unittest

from aiohttp import ClientError, web

import crawling


@contextmanager
def capture_logging():
    """See test_redirect for usage."""
    logger = logging.getLogger('crawling')
    level = logger.level
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(io.StringIO())
    logger.addHandler(handler)

    class Messages:
        def __contains__(self, item):
            return item in handler.stream.getvalue()

        def __repr__(self):
            return repr(handler.stream.getvalue())

    try:
        yield Messages()
    finally:
        logger.removeHandler(handler)
        logger.setLevel(level)


class TestCrawler(unittest.TestCase):
    # Jesse gratefully copied some of this from asyncio's and aiohttp's tests.

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.addCleanup(self.loop.close)
        self.port = self._find_unused_port()
        self.app_url = "http://127.0.0.1:{}".format(self.port)
        self.app = self.loop.run_until_complete(self._create_server())
        self.crawler = None

    def _find_unused_port(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('127.0.0.1', 0))
        port = s.getsockname()[1]
        s.close()
        return port

    @asyncio.coroutine
    def _create_server(self):
        app = web.Application(loop=self.loop)
        handler_factory = app.make_handler(debug=True)
        srv = yield from self.loop.create_server(
            handler_factory, '127.0.0.1', self.port)

        # Prevent "Task destroyed but it is pending" warnings.
        self.addCleanup(lambda: self.loop.run_until_complete(
            handler_factory.finish_connections()))

        self.addCleanup(srv.close)
        return app

    def add_handler(self, url, handler):
        self.app.router.add_route('GET', url, handler)

    def add_page(self, url='/', links=None, body=None, content_type=None):
        if not body:
            text = ''.join('<a href="{}"></a>'.format(link)
                           for link in links or [])
            body = text.encode('utf-8')

        if content_type is None:
            content_type = 'text/html; charset=utf-8'

        @asyncio.coroutine
        def handler(req):
            yield from req.read()
            return web.Response(body=body, headers=[
                ('CONTENT-TYPE', content_type)])

        self.add_handler(url, handler)
        return self.app_url + url

    def add_redirect(self, url, link):
        @asyncio.coroutine
        def handler(_):
            raise web.HTTPFound(link)

        self.add_handler(url, handler)
        return self.app_url + url

    def assertDoneCount(self, n):
        self.assertEqual(n, len(self.crawler.done),
                         "Expected {} URLs done, got {}".format(
                             n, len(self.crawler.done)))

    def assertStat(self, stat_index=0, **kwargs):
        stat = self.crawler.done[stat_index]
        for name, value in kwargs.items():
            msg = '{}.{} not equal to {!r}'.format(stat, name, value)
            self.assertEqual(getattr(stat, name), value, msg)

    def crawl(self, urls=None, *args, **kwargs):
        if self.crawler:
            self.crawler.close()
        if urls is None:
            urls = [self.app_url]
        self.crawler = crawling.Crawler(urls, *args, loop=self.loop, **kwargs)
        self.addCleanup(self.crawler.close)
        self.loop.run_until_complete(self.crawler.crawl())

    def test_link(self):
        # "/" links to foo, which is missing.
        self.add_page('/', ['/foo'])
        self.crawl()
        self.assertDoneCount(2)
        self.assertStat(url=self.app_url,
                        num_urls=1,
                        num_new_urls=1)

        self.assertStat(1, url=self.app_url + '/foo', status=404)

    def test_link_cycle(self):
        # foo and bar link to each other.
        url = self.add_page('/foo', ['/bar'])
        self.add_page('/bar', ['/foo'])
        self.crawl([url])
        self.assertDoneCount(2)
        self.assertStat(url=self.app_url + '/foo',
                        num_urls=1,
                        num_new_urls=1)

        self.assertStat(1,
                        url=self.app_url + '/bar',
                        num_urls=1,
                        num_new_urls=0)

    def test_prohibited_host(self):
        # Link to example.com.
        self.add_page('/', ['http://example.com'])
        self.crawl()
        self.assertStat(num_urls=0)

    def test_strict_host_checking(self):
        crawler = crawling.Crawler(['http://example.com'], loop=self.loop)
        self.assertTrue(crawler.url_allowed("http://www.example.com"))
        self.assertFalse(crawler.url_allowed("http://foo.example.com"))

    def test_lenient_host_checking(self):
        crawler = crawling.Crawler(['http://example.com'], strict=False,
                                   loop=self.loop)
        self.assertTrue(crawler.url_allowed("http://www.example.com"))
        self.assertTrue(crawler.url_allowed("http://foo.example.com"))

    def test_exclude(self):
        crawler = crawling.Crawler(['http://example.com'],
                                   exclude=r'.*pattern', loop=self.loop)
        self.assertTrue(crawler.url_allowed("http://example.com"))
        self.assertFalse(crawler.url_allowed("http://example.com/pattern"))

    def test_roots(self):
        crawler = crawling.Crawler(['http://a', 'http://b', 'not-a-host'],
                                   loop=self.loop)
        self.assertTrue(crawler.url_allowed("http://a/a"))
        self.assertTrue(crawler.url_allowed("http://b/b"))
        self.assertFalse(crawler.url_allowed("http://c/c"))
        self.assertFalse(crawler.url_allowed("http://127.0.0.1"))

    def test_deep_root(self):
        # Make sure 'a' is a root domain if the root is a link deep in 'a'.
        crawler = crawling.Crawler(['http://a/a#fragment'], loop=self.loop)
        self.assertTrue(crawler.url_allowed("http://a/b"))

    def test_redirect(self):
        # "/" redirects to "/foo", and "/foo" redirects to "/bar".
        foo = self.app_url + '/foo'
        bar = self.app_url + '/bar'

        url = self.add_redirect('/', foo)
        self.add_redirect('/foo', bar)
        self.crawl([url])
        self.assertStat(0, status=302, next_url=foo)
        self.assertStat(1, status=302, next_url=bar)
        self.assertStat(2, status=404)

        with capture_logging() as messages:
            self.crawl([url], max_redirect=1)
            self.assertDoneCount(2)
            self.assertStat(status=302, next_url=foo)

        self.assertIn('redirect limit reached', messages)

    def test_redirect_cycle(self):
        foo = self.app_url + '/foo'
        bar = self.app_url + '/bar'

        url = self.add_redirect('/bar', foo)
        self.add_redirect('/foo', bar)
        self.crawl([url])
        self.assertStat(0, status=302, next_url=foo)
        self.assertStat(1, status=302, next_url=bar)
        self.assertDoneCount(2)

    def test_redirect_join(self):
        # Set up redirects:
        #   foo -> baz
        #   bar -> baz -> quux
        foo = self.app_url + '/foo'
        bar = self.app_url + '/bar'
        baz = self.app_url + '/baz'
        quux = self.app_url + '/quux'

        self.add_redirect('/foo', baz)
        self.add_redirect('/bar', baz)
        self.add_redirect('/baz', quux)

        # Start crawling foo and bar. We follow the foo -> baz redirect but
        # not bar -> baz, since by then baz is already seen.
        self.crawl([foo, bar])
        import pprint
        pprint.pprint(self.crawler.done)
        self.assertStat(0, url=foo, status=302, next_url=baz)

        # We fetched bar and saw it redirected to baz.
        self.assertStat(1, url=bar, status=302, next_url=baz)

        # But we only fetched baz once.
        self.assertStat(2, url=baz, status=302, next_url=quux)
        self.assertStat(3, url=quux, status=404)
        self.assertDoneCount(4)

    def test_max_tasks(self):
        n_tasks = 0
        max_tasks = 0

        @asyncio.coroutine
        def handler(_):
            nonlocal n_tasks, max_tasks
            n_tasks += 1
            max_tasks = max(n_tasks, max_tasks)
            yield from asyncio.sleep(0.01, loop=self.loop)
            n_tasks -= 1
            return web.Response(body=b'')

        urls = ['/0', '/1', '/2']
        for url in urls:
            self.add_handler(url, handler)
        home = self.add_page('/', urls)
        self.crawl([home], max_tasks=2)
        self.assertEqual(2, max_tasks)
        self.assertDoneCount(4)

    def test_max_tries(self):
        n_tries = 0

        @asyncio.coroutine
        def handler(req):
            nonlocal n_tries
            n_tries += 1
            if n_tries <= 2:
                req.transport.close()  # Network failure.
            return web.Response(body=b'')

        self.add_handler('/', handler)
        with capture_logging() as messages:
            self.crawl([self.app_url])
        self.assertDoneCount(1)
        self.assertStat(status=200)
        self.assertIn('try 1', messages)
        self.assertIn('try 2', messages)

        n_tries = 0
        with capture_logging() as messages:
            self.crawl([self.app_url], max_tries=1)
        self.assertDoneCount(1)
        self.assertStat(status=None)
        stat = self.crawler.done[0]
        self.assertIsInstance(stat.exception, ClientError)
        self.assertIn('failed after 1 tries', messages)

    def test_encoding(self):
        def test_charset(charset, encoding):
            if charset:
                content_type = 'text/html; charset={}'.format(charset)
            else:
                content_type = 'text/html'
            url = '/' + charset
            self.add_page(url, content_type=content_type)
            self.crawl([self.app_url + url])
            self.assertStat(encoding=encoding)

        test_charset('', 'utf-8')
        test_charset('utf-8', 'utf-8')
        test_charset('ascii', 'ascii')

    def test_content_type(self):
        self.add_page(content_type='foo')
        self.crawl([self.app_url])
        self.assertStat(content_type='foo')

    def test_non_html(self):
        # Should search only XML and HTML for links, not other content types.
        body = ('<a href="{}">'.format(self.app_url)).encode('utf-8')

        self.add_page('/xml', body=body, content_type='application/xml')
        self.crawl([self.app_url + '/xml'])
        self.assertStat(0, content_type='application/xml', num_urls=1)
        self.assertStat(1, url=self.app_url)

        self.add_page('/image', content_type='image')
        self.crawl([self.app_url + '/image'])
        self.assertStat(content_type='image', num_urls=0)

    def test_non_http(self):
        body = '<a href="ftp://example.com">'.encode('utf-8')
        self.add_page(body=body)
        self.crawl()
        self.assertStat(num_urls=0)


if __name__ == '__main__':
    unittest.main()



DEFAULT_CRAWL_CONFIG = {
    # Required
    'seed_urls' : [
    
    ],
    
    'collections_path' : '',

    'collection_name' : '',

    # Optional
    
    'crawl_filter_regexps' : [
                        
    ],

    'crawl_required_regexps' : [
    ],        
            
    'index_filter_regexps' : [
        
    ],
    
    'index_required_regexps' : [
                          
    ], 

    'max_redirects_per_url' : 10, 
    
    'max_tries_per_url' : 1, 
    
    'max_workers' : 5, 
    
    'max_saved_responses' : 20, 

    'login_url' : '',
    'login_data' : {},
    
    # Unimplemented 
    'max_hops' : 10
}

PRODUCTION_CRAWL_CONFIG = DEFAULT_CRAWL_CONFIG.copy()

PRODUCTION_CRAWL_CONFIG.update({
    'max_saved_responses' : 1000000000, 
    'max_hops' : 100,
    'max_workers' : 100
})



"""Reporting subsystem for web crawler."""

import time


class Stats:
    """Record stats of various sorts."""

    def __init__(self):
        self.stats = {}

    def add(self, key, count=1):
        self.stats[key] = self.stats.get(key, 0) + count

    def report(self, file=None):
        for key, count in sorted(self.stats.items()):
            print('%10d' % count, key, file=file)


def report(crawler, file=None):
    """Print a report on all completed URLs."""
    t1 = crawler.t1 or time.time()
    dt = t1 - crawler.t0
    if dt and crawler.max_workers:
        speed = len(crawler.done) / dt / crawler.max_workers
    else:
        speed = 0
    stats = Stats()
    print('*** Report ***', file=file)
    try:
        show = list(crawler.done)
        show.sort(key=lambda _stat: _stat.url)
        for stat in show:
            url_report(stat, stats, file=file)
    except KeyboardInterrupt:
        print('\nInterrupted', file=file)
    print('Finished', len(crawler.done),
          'urls in %.3f secs' % dt,
          '(max_workers=%d)' % crawler.max_workers,
          '(%.3f urls/sec/task)' % speed,
          file=file)
    stats.report(file=file)
    print('Todo:', crawler.q.qsize(), file=file)
    print('Done:', len(crawler.done), file=file)
    print('Date:', time.ctime(), 'local time', file=file)


def url_report(stat, stats, file=None):
    """Print a report on the state for this URL.

    Also update the Stats instance.
    """
    if stat.exception:
        stats.add('fail')
        stats.add('fail_' + str(stat.exception.__class__.__name__))
        print(stat.url, 'error', stat.exception, file=file)
    elif stat.next_url:
        stats.add('redirect')
        print(stat.url, stat.status, 'redirect', stat.next_url,
              file=file)
    elif stat.content_type == 'text/html':
        stats.add('html')
        stats.add('html_bytes', stat.size)
        print(stat.url, stat.status,
              stat.content_type, stat.encoding,
              stat.size,
              '%d/%d' % (stat.num_new_urls, stat.num_urls),
              file=file)
    else:
        if stat.status == 200:
            stats.add('other')
            stats.add('other_bytes', stat.size)
        else:
            stats.add('error')
            stats.add('error_bytes', stat.size)
            stats.add('status_%s' % stat.status)
        print(stat.url, stat.status,
              stat.content_type, stat.encoding,
              stat.size,
              file=file)



import os

try:
    from ZODB.serialize import referencesf
    from ZODB.DB import DB
    import transaction
    from BTrees.OOBTree import OOBTree
    import time
except NameError:
    msg = 'ZODB not properly installed and cannot be used as data backend.\n'
    msg += 'Use `pip3 install ZODB zodbpickle` to install, or use a different backend.'
    print(msg)


class CrawlService():

    def __init__(self, project_name, storage_object, crawl_plugin_class):
        self.project_name = project_name
        self.storage_object = storage_object
        self.plugins = {}
        self.server = None
        self.crawl_plugin_class = crawl_plugin_class
        self.get_server()
        self.get_crawl_plugins()

    def get_server(self):
        raise NotImplementedError("get_server not implemented")

    def get_crawl_plugins(self):
        raise NotImplementedError("get_crawl_plugins not implemented")

    def get_crawl_plugin(self, plugin_name):
        return self.crawl_plugin_class(self.project_name, self.server, plugin_name)

    def run(self, plugin_name):
        cplug = self.get_crawl_plugin(plugin_name)
        cplug.run()

    def run_all(self):
        for plugin in self.plugins:
            self.run(plugin)


class CrawlFileService(CrawlService):

    def get_server(self):
        root = self.storage_object['path']
        tps = ['plugins', 'documents', 'template_dict']
        self.server = {tp: os.path.join(root, self.project_name + '-crawler-' + tp)
                       for tp in tps}
        # create paths if they don't exist
        for paths in self.server.values():
            os.makedirs(paths, exist_ok=True)

    def get_crawl_plugins(self):
        self.plugins = {}
        for fn in os.listdir(self.server['plugins']):
            if fn != 'default':
                with open(os.path.join(self.server['plugins'], fn)) as f:
                    # not yet parsed config, not sure if that is a problem
                    self.plugins[fn] = f.read()


class CrawlCloudantService(CrawlService):

    def get_server(self):
        account = self.storage_object

        # create dbs if they don't exist
        account.database(self.project_name + '-crawler-plugins').put()
        account.database(self.project_name + '-crawler-documents').put()
        account.database(self.project_name + '-crawler-template_dict').put()

        self.server = account

    def get_crawl_plugins(self):
        plugin_db = self.server.database(self.project_name + '-crawler-plugins')
        db_uri = '{}/_all_docs?include_docs=true'.format(plugin_db.uri)
        service_rows = plugin_db.get(db_uri).json()['rows']
        self.plugins = {row['doc']['plugin_name']: row['doc'] for row in service_rows
                        if 'default' != row['doc']['plugin_name']}


class CrawlZODBService(CrawlService):

    def get_server(self):
        # In services, a FileStorage (or perhaps other backend) object has to be provided
        db = DB(self.storage_object)
        connection = db.open()
        self.server = connection.root()

        # create dbs if they don't exist
        tables = ['plugins', 'documents', 'template-dict']
        for table in tables:
            if table not in self.server:
                self.server[table] = OOBTree()
                transaction.commit()

    def pack(self):
        self.storage_object.pack(time.time(), referencesf)

    def get_crawl_plugins(self):
        return {k: doc for k, doc in self.server['plugins'].items() if k != 'default'}


class CrawlElasticSearchService(CrawlService):

    def create_index_if_not_existent(self, name, request_body=None):
        if not self.server.indices.exists(name):
            if request_body is None:
                request_body = {
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 0
                    }
                }
            self.server.indices.create(index=name, body=request_body)

    def get_server(self):
        self.server = self.storage_object

        # create dbs if they don't exist
        for db in ['plugins', 'documents', 'template_dict']:
            self.create_index_if_not_existent(self.project_name + '-crawler-' + db)

    def get_crawl_plugins(self):
        query = {"query": {"match_all": {}}}
        res = self.server.search(body=query, doc_type='plugin',
                                 index=self.project_name + "-crawler-plugins")
        return {doc['_id']: doc for doc in res['hits']['hits']
                if doc['_id'] != 'default'}



from setuptools import setup

MAJOR_VERSION = '0'
MINOR_VERSION = '0'
MICRO_VERSION = '114'
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(name='sky',
      version=VERSION,
      description='AI powered scraping in Python 3',
      url='https://github.com/kootenpv/sky',
      author='Pascal van Kooten',
      author_email='kootenpv@gmail.com',
      license='MIT',
      packages=['sky', 'sky/crawler', 'sky/view'],
      install_requires=[
          'distribute', 'lxml', 'tldextract', 'requests', 'justext', 'langdetect',
          'python-dateutil', 'sh', 'beautifulsoup4', 'aiohttp', 'tornado'
      ],
      entry_points={
          'console_scripts': ['sky = sky.sky:main']
      },
      # optional: ZODB, zodbpickle, cloudant, elasticsearch, selenium, asciitree, nltk
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Customer Service',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
          'Operating System :: Microsoft',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Unix',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Topic :: Software Development',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Software Development :: Debuggers',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Software Distribution',
          'Topic :: System :: Systems Administration',
          'Topic :: Utilities'
      ],
      include_package_data=True,
      zip_safe=False,
      platforms='any')



#!/usr/bin/env python3.4

"""A simple web crawler -- main driver program."""

import asyncio
import logging
import sys
import os

from sky.crawler.crawling import Crawler
from sky.crawler.reporting import report

def get_loop(select = False):
    if os.name == 'nt':
        from asyncio.windows_events import ProactorEventLoop
        loop = ProactorEventLoop()
    elif select:
        loop = asyncio.SelectorEventLoop()
    else:
        loop = asyncio.get_event_loop()
    return loop

def get_config(config, loop): 
    for i in range(len(config['seed_urls'])):
        if '://' not in config['seed_urls'][i]:
            config['seed_urls'][i] = 'http://' + config['seed_urls'][i]

    config['loop'] = loop
    
    return config
    
    
def start(config, crawler_class = Crawler, save_data_fn = None, logging_level = 2):
    """Main program.

    Parse arguments, set up event loop, run crawler, print report.
    """

    logging_levels = [logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
    logging.basicConfig(level = logging_levels[min(logging_level, len(logging_levels)-1)])

    loop = asyncio.SelectorEventLoop()

    asyncio.set_event_loop(loop)
    conf = get_config(config, loop) 

    crawler = crawler_class(conf) 

    if save_data_fn is not None:
        crawler.save_data = save_data_fn

    if crawler.login_url:
        loop.run_until_complete(crawler.login())
    
    try:
        loop.run_until_complete(crawler.crawl())  # Crawler gonna crawl.
    except KeyboardInterrupt:
        sys.stderr.flush()
        print('\nInterrupted\n')
    finally:
        result = crawler.finish_leftovers()
        report(crawler) 
        crawler.close()
        loop.close()
    return result
